/* ----------------------------------------------------------------------------
Question Server
 GitLab: <https://gitlab.com/MrFry/mrfrys-node-server>

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program. If not, see <https://www.gnu.org/licenses/>.

 ------------------------------------------------------------------------- */

module.exports = {
  ProcessIncomingRequest: ProcessIncomingRequest,
  LoadJSON: LoadJSON,
  backupData: backupData,
}

const recDataFile = './stats/recdata'
const dataLockFile = './data/lockData'

import logger from '../utils/logger.js'
import { searchData, createQuestion } from '../utils/classes.js'
import idStats from '../utils/ids.js'
import utils from '../utils/utils.js'
import { addQuestion, getSubjNameWithoutYear } from './classes.js'

// types
import { QuestionDb, Question, User } from '../types/basicTypes'
import { DataFile } from '../modules/api/api'

// if a recievend question doesnt match at least this % to any other question in the db it gets
// added to db
const minMatchToAmmountToAdd = 90

const writeAfter = 1 // write after # of adds FIXME: set reasonable save rate
let currWrites = 0

interface RecievedData {
  quiz: Array<Question>
  subj: string
  id: string
  version: string
  scriptVersion: string
}

function ProcessIncomingRequest(
  recievedData: RecievedData,
  questionDbs: Array<QuestionDb>,
  dryRun: boolean,
  user: User
) {
  return Promise.all(
    questionDbs.map((qdb) => {
      return ProcessIncomingRequestUsingDb(recievedData, qdb, dryRun, user)
    })
  )
}

function ProcessIncomingRequestUsingDb(
  recievedData: RecievedData,
  qdb: QuestionDb,
  dryRun: boolean,
  user: User
) {
  return new Promise((resolve, reject) => {
    logger.DebugLog('Processing incoming request', 'actions', 1)

    if (recievedData === undefined) {
      logger.Log('\tRecieved data is undefined!', logger.GetColor('redbg'))
      reject(new Error('Recieved data is undefined!'))
    }

    try {
      let towrite = logger.GetDateString() + '\n'
      towrite +=
        '------------------------------------------------------------------------------\n'
      if (typeof recievedData === 'object') {
        towrite += JSON.stringify(recievedData)
      } else {
        towrite += recievedData
      }
      towrite +=
        '\n------------------------------------------------------------------------------\n'
      utils.AppendToFile(towrite, recDataFile)
      logger.DebugLog('recDataFile written', 'actions', 1)
    } catch (err) {
      logger.log('Error writing recieved data.')
    }

    if (utils.FileExists(dataLockFile)) {
      logger.Log(
        'Data lock file exists, skipping recieved data processing',
        logger.GetColor('red')
      )
      resolve(-1)
      return
    }

    try {
      // recievedData: { version: "", id: "", subj: "" quiz: {} }
      const recievedQuestions = []

      logger.DebugLog('recievedData JSON  parsed', 'actions', 1)
      logger.DebugLog(recievedData, 'actions', 3)
      const allQLength = recievedData.quiz.length
      const questionSearchPromises = []
      recievedData.quiz.forEach((question) => {
        logger.DebugLog('Question:', 'actions', 2)
        logger.DebugLog(question, 'actions', 2)
        const currentQuestion = createQuestion(
          question.Q,
          question.A,
          question.data
        )
        logger.DebugLog(
          'Searching for question in subj ' + recievedData.subj,
          'actions',
          3
        )
        logger.DebugLog(currentQuestion, 'actions', 3)
        recievedQuestions.push(currentQuestion)
        questionSearchPromises.push(
          searchData(qdb.data, currentQuestion, recievedData.subj)
        )
      })

      Promise.all(questionSearchPromises)
        .then((results) => {
          const allQuestions = [] // all new questions here that do not have result
          results.forEach((result, i) => {
            const add = result.every((res) => {
              return res.match < minMatchToAmmountToAdd
            })
            if (add) {
              allQuestions.push(recievedQuestions[i])
            }
          })

          try {
            let color = logger.GetColor('green')
            let msg = ''
            if (allQuestions.length > 0) {
              color = logger.GetColor('blue')
              msg += `New questions: ${allQuestions.length} ( All: ${allQLength} )`
              allQuestions.forEach((currentQuestion) => {
                const sName = getSubjNameWithoutYear(recievedData.subj)
                logger.DebugLog(
                  'Adding question with subjName: ' + sName + ' :',
                  'actions',
                  3
                )
                logger.DebugLog(currentQuestion, 'actions', 3)
                addQuestion(qdb.data, sName, currentQuestion)
              })

              currWrites++
              logger.DebugLog(
                'currWrites for data.json: ' + currWrites,
                'actions',
                1
              )
              if (currWrites >= writeAfter && !dryRun) {
                currWrites = 0
                logger.DebugLog('Writing data.json', 'actions', 1)
                utils.WriteFile(JSON.stringify(qdb.data), qdb.path)
                logger.Log('\tData file written', color)
              } else if (dryRun) {
                logger.Log('\tDry run')
              }
            } else {
              msg += `No new data ( ${allQLength} )`
            }

            let subjRow = '\t' + recievedData.subj
            if (recievedData.id) {
              subjRow += ' ( CID: ' + logger.logHashed(recievedData.id) + ')'
            }
            idStats.LogId(
              user.id,
              recievedData.subj,
              allQuestions.length,
              allQLength
            )
            logger.Log(subjRow)
            if (recievedData.version !== undefined) {
              msg += '. Version: ' + recievedData.version
            }

            logger.Log('\t' + msg, color)
            logger.DebugLog('New Questions:', 'actions', 2)
            logger.DebugLog(allQuestions, 'actions', 2)

            logger.DebugLog('ProcessIncomingRequest done', 'actions', 1)
            resolve(allQuestions.length)
          } catch (error) {
            console.error(error)
            logger.Log(
              'Error while processing processData worker result!',
              logger.GetColor('redbg')
            )
            reject(
              new Error('Error while processing processData worker result!')
            )
          }
        })
        .catch((err) => {
          logger.Log(
            'Error while searching for questions in ProcessIncomingRequest!',
            logger.GetColor('redbg')
          )
          console.error(err)
        })
    } catch (err) {
      console.error(err)
      logger.Log(
        'There was en error handling incoming quiz data, see stderr',
        logger.GetColor('redbg')
      )
      reject(new Error('Couldnt parse JSON data'))
    }
  })
}

function LoadJSON(dataFiles: Array<DataFile>) {
  return dataFiles.reduce((acc, dataFile) => {
    if (!utils.FileExists(dataFile.path)) {
      utils.WriteFile(JSON.stringify([]), dataFile.path)
    }

    try {
      acc.push({
        ...dataFile,
        data: JSON.parse(utils.ReadFile(dataFile.path)),
      })
    } catch (err) {
      console.error(err)
      logger.Log(
        "data is undefined! Couldn't load data!",
        logger.GetColor('redbg')
      )
    }
    return acc
  }, [])
}

function backupData(questionDbs: Array<DataFile>) {
  questionDbs.forEach((data) => {
    const path = './publicDirs/qminingPublic/backs/'
    utils.CreatePath(path)
    try {
      logger.Log(`Backing up ${data.name}...`)
      utils.WriteFile(
        JSON.stringify(data.data),
        `${path}${data.name}_${utils.GetDateString(true)}.json`
      )
      logger.Log('Done')
    } catch (err) {
      logger.Log(
        `Error backing up data file ${data.name}!`,
        logger.GetColor('redbg')
      )
      console.error(err)
    }
  })
}
